/* ////////////////////////////////////////////////////////////

File Name: workorder_constedge.cpp
Copyright (c) 2015 Soji Yamakawa.  All rights reserved.
http://www.ysflight.com

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, 
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, 
   this list of conditions and the following disclaimer in the documentation 
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

//////////////////////////////////////////////////////////// */

#include "workorder.h"
#include <ysshellextedit_constedgeutil.h>


YSRESULT GeblCmd_WorkOrder::RunConstEdgeWorkOrder(const YsString &workOrder,const YsArray <YsString,16> &args)
{
	if(2<=args.GetN())
	{
		YsString subCmd=args[1];
		subCmd.Capitalize();
		if(0==strcmp(args[1],"CLEAR"))
		{
			return RunConstEdgeClear(workOrder,args);
		}
		else if(0==strcmp(args[1],"ADD"))
		{
			return RunConstEdgeAdd(workOrder,args);
		}
		else if(0==strcmp(args[1],"RECONSIDER"))
		{
			return RunConstEdgeReconsider(workOrder,args);
		}

		YsString errorReason;
		errorReason.Printf("Unrecognized sub command [%s]",args[1].Txt());
		ShowError(workOrder,errorReason);
	}
	else
	{
		ShowError(workOrder,"Sub-command not given.");
	}
	return YSERR;
}

YSRESULT GeblCmd_WorkOrder::RunConstEdgeClear(const YsString &,const YsArray <YsString,16> &)
{
	auto &shl=*slHd;
	for(auto ceHd : shl.AllConstEdge())
	{
		shl.DeleteConstEdge(ceHd);
	}
	return YSOK;
}

YSRESULT GeblCmd_WorkOrder::RunConstEdgeAdd(const YsString &workOrder,const YsArray <YsString,16> &args)
{
	YsString option=args[2];
	option.Capitalize();
	if(0==strcmp(option,"FGBOUNDARY"))
	{
		return YsShellExtEdit_AddConstEdgeAlongFaceGroupBoundary(*slHd);
	}
	else if(0==strcmp(option,"DHA"))
	{
		if(4<=args.GetN())
		{
			const double angle=YsDegToRad(atof(args[3]));
			return YsShellExtEdit_AddConstEdgeAlongFaceGroupBoundary(*slHd);
		}
		else
		{
			ShowError(workOrder,"Too few arguments.");
			return YSERR;
		}
	}
	else if(0==strcmp(option,"NONMANIFOLD"))
	{
		YsShellExtEdit_AddConstEdgeAlongNonManifoldEdge(*slHd);
	}

	YsString errorReason;
	errorReason.Printf("Not supported or not implemented yet. [%s]\n",option.Txt());
	ShowError(workOrder,errorReason);
	return YSERR;
}

YSRESULT GeblCmd_WorkOrder::RunConstEdgeReconsider(const YsString &workOrder,const YsArray <YsString,16> &args)
{
	YsShellExtEdit_ReconsiderConstEdge(*slHd);
	return YSOK;
}
